/* 
 * The contents of this file are subject to the GrammarSoft Public
 * License Version 1.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.grammarsoft.com/GSPL
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is the VISL CG Parser.
 * 
 * The Initial Developer of the Original Code is Martin Carlsen, 
 * GrammarSoft, and SDU Odense University.  
 * Portions created by Martin Carlsen, Grammarsoft, and SDU Odense 
 * University are Copyright (C) 1999-2000 Martin Carlsen, GrammarSoft, 
 * and SDU Odense University.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * Alternatively, the contents of this file may be used under the
 * terms of the GNU General Public License Version 2 or later (the
 * "GPL"), in which case the provisions of the GPL are applicable 
 * instead of those above.  If you wish to allow use of your 
 * version of this file only under the terms of the GPL and not to
 * allow others to use your version of this file under the GSPL,
 * indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by
 * the GPL.  If you do not delete the provisions above, a recipient
 * may use your version of this file under either the GSPL or the
 * GPL.
 */
#ifndef _Set_hh_
#define _Set_hh_

/**
 * Pure virtual Set
 */
class Set
{
public:
  virtual bool targets(Combined * tag) const = 0;
  virtual bool matches(Combined * tag) const = 0;
  virtual bool matches(Cohort * cohort, bool careful=false) const = 0;
  virtual ostream& put(ostream& os) const = 0;
  virtual void accept(Visitor * v) const = 0;
  virtual ~Set(){};
};

/**
 * This is a reference to a set.
 * Either to an anonymous set (SetUnion, SetDifference, SetConcatenation)
 * or a named, previously declared set (SetDeclaration, ListDeclaration)
 */
class SetReference : public Set
{ 
private:
  /** The declaration of the set referred to */
  Set * decl;
  /** The name of the set referred to (optional) */
  string setname;
  
public:
  Set * getDeclaration() const { return decl; }
  string getName() const { return setname; }
  
  virtual bool targets(Combined * tag) const;
  virtual bool matches(Combined * tag) const;
  virtual bool matches(Cohort * cohort, bool careful=false) const;
  virtual void accept(Visitor * v) const { v->visitSetReference(this); }
  virtual ostream& put(ostream& os) const;
  
  SetReference(Set * s=0, const string& name = *empty_string)
    : decl(s), setname(name)
    {
    }
  virtual ~SetReference() { }
};

/**
 * A named set.
 */
class SetDeclaration : public Set
{ 
private:
  /** The set expression defining the set */ 
  Set * expr;
  /** The name of the set */ 
  string setname;
  
public:
  Set * getExpr() const { return expr; }
  string getName() const { return setname; }
  
  virtual bool targets(Combined * tag) const;
  virtual bool matches(Combined * tag) const;
  virtual bool matches(Cohort * cohort, bool careful=false) const;
  virtual void accept(Visitor * v) const { v->visitSetDeclaration(this); }
  virtual ostream& put(ostream& os) const;
  
  SetDeclaration(Set * xpr, const string& name = *empty_string)
    : expr(xpr), setname(name)
    {
    }
  
  virtual ~SetDeclaration() { }
};

/**
 * A named List.
 */
class ListDeclaration : public Set
{ 
private:
  /** The list of combined tags */
  list<Combined*>* combinedTags;
  /** The name of the set */
  string listname;
  
public:
  list<Combined*> * getTags() const { return combinedTags; }
  string getName() const { return listname; }
  
  virtual bool targets(Combined * tag) const;
  virtual bool matches(Combined * tag) const;
  virtual bool matches(Cohort * cohort, bool careful=false) const;
  virtual void accept(Visitor * v) const {v->visitListDeclaration(this); }
  virtual ostream& put(ostream& os) const;
  
  ListDeclaration(list<Combined*>* combtags=0, const string & name=*empty_string)
    : combinedTags(combtags), listname(name)
    {
    }
  virtual ~ListDeclaration() { }
};

/**
 * Union of two sets: set1 OR set2
 */
class SetUnion : public Set
{ 
private:
  Set * set1;
  Set * set2;
  
public:
  Set * getSet1() const { return set1; }
  Set * getSet2() const { return set2; }
  
  virtual bool targets(Combined * tag) const;
  virtual bool matches(Combined * tag) const;
  virtual bool matches(Cohort * cohort, bool careful=false) const;
  virtual void accept(Visitor * v) const {v->visitSetUnion(this); }
  virtual ostream& put(ostream& os) const;
  
  SetUnion(Set * a, Set * b)
    : set1(a), set2(b)
    {
    }
  virtual ~SetUnion() { }
};

/**
 * Intersection of two sets: set1 + set2
 */
class SetConcatenation : public Set
{
private:
  Set * set1;
  Set * set2;
  
public:
  Set * getSet1() const { return set1; }
  Set * getSet2() const { return set2; }
  
  virtual bool targets(Combined * tag) const;
  virtual bool matches(Combined * tag) const;
  virtual bool matches(Cohort * cohort, bool careful=false) const;
  virtual void accept(Visitor * v) const { v->visitSetConcatenation(this); }
  virtual ostream& put(ostream& os) const;
  
  SetConcatenation(Set * a, Set * b)
    : set1(a), set2(b)
    {
    }
  virtual ~SetConcatenation() { }
};

/**
 * Difference of two sets: set1 - set2
 */
class SetDifference : public Set
{
private:
  Set * set1;
  Set * set2;
  
public:
  Set * getSet1() const { return set1; }
  Set * getSet2() const { return set2; }
  
  virtual bool targets(Combined * tag) const;
  virtual bool matches(Combined * tag) const;
  virtual bool matches(Cohort * cohort, bool careful=false) const;
  virtual void accept(Visitor * v) const { v->visitSetDifference(this); }
  virtual ostream& put(ostream& os) const;
  
  SetDifference(Set * a, Set * b)
    : set1(a), set2(b)
    {
    }
  virtual ~SetDifference() { }
};

#endif
