/* 
 * The contents of this file are subject to the GrammarSoft Public
 * License Version 1.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.grammarsoft.com/GSPL
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is the VISL CG Parser.
 * 
 * The Initial Developer of the Original Code is Martin Carlsen, 
 * GrammarSoft, and SDU Odense University.  
 * Portions created by Martin Carlsen, Grammarsoft, and SDU Odense 
 * University are Copyright (C) 1999-2000 Martin Carlsen, GrammarSoft, 
 * and SDU Odense University.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * Alternatively, the contents of this file may be used under the
 * terms of the GNU General Public License Version 2 or later (the
 * "GPL"), in which case the provisions of the GPL are applicable 
 * instead of those above.  If you wish to allow use of your 
 * version of this file only under the terms of the GPL and not to
 * allow others to use your version of this file under the GSPL,
 * indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by
 * the GPL.  If you do not delete the provisions above, a recipient
 * may use your version of this file under either the GSPL or the
 * GPL.
 */
#include <map>
#include <list>
#include <vector>
#include <string>
#include <stdio.h>
#include <hash_map>
#include <algorithm>
#include <functional>

#include "cg.data_structures.hh"
#include "Debug.hh"
#include "CGCompiler.hh"

/**
 * integer to decimal string.
 */
string * itos(unsigned int i)
{
  if(i==0) return new string("0");
  string s = "";
  while(i>0)
    {
      char digit = '0' + i%10;
      i = i/10;
      s = digit + s;
    }
  return new string(s);
}

/**
 * integer to hexadecimal string
 */
string * xtos(unsigned int i)
{
  if(i==0) return new string("0x0");
  string s = "";
  while(i>0)
    {
      unsigned int n = i%16;
      char digit = (n>9?'a'+n-10:'0'+n);
      i = i/16;
      s = digit + s;
    }
  return new string("0x"+s);
}


/**
 * Apply a correction rule to a reading or cohort.
 */
bool Correction::correct(vector<Cohort*>* cohorts, int pos, list<Combined*>::iterator ri) const
{
  Cohort * cohort = (*cohorts)[pos];
  Combined * reading = *ri;
  
  // Does the wordform match?
  if(! (wordform==*empty_string || wordform==cohort->getWordform())) return false;
  
  // Does the target match? (for SUBSTITUTE only)
  if(target != 0 && ! target->matches(reading) ) return false;
  
  // Do all context tests match?
  list<Context*>::iterator a = tests->begin();
  for(list<Context*>::iterator
	i = a,
	z = tests->end();
      i!=z; ++i)
    {
      if(! (*i)->matchCohort(cohorts,pos))
	{
	  // Move non-matching tests to the front, 
	  // assuming that they will fail more often than other tests.
	  // Thus most often the _first_ constraint tested will
	  // be the one that fails (saving some time)
	  iter_swap(a,i);
	  
	  return false;
	}
    }
  
  Combined * toAppend = new Combined();
  for(Combined::iterator
	i = insertion->begin(),
	z = insertion->end();
      i!=z; ++i)
    {
      toAppend->push_back(*i);
    }
  
  // Carry out the operation
  switch(op)
    {
    case op_append:
      cohort->getReadings()->push_front(toAppend);
      return true;
      
    case op_substitute:
#ifndef BINARY
      // Debug mode
      if(Debug::instance()->debug) {
	if(reading->has(CGCompiler::correctTag()))
	  {
	    cout << "WARNING: ";
	    reading->put(cout);
	    cout << (Debug::instance()->debug_cg2 ?
		     " may be modified by " : 
		     " modified by ");
	    cout << (*this) << "\n";
	  }
	if(Debug::instance()->debug_cg2) return false;
      }
#endif
  
      Combined::iterator
	i = reading->begin(),
	z = reading->end(),
	ri = removal->begin(),
	rz = removal->end(),
	a = i;
      bool doInsertion = false;
      // remove all tags in the removal list from the reading 
      // (if they exist in the reading, that is).
      while(i!=z && ri!=rz)
	{
	  if((*i)==(*ri)){
	    Combined::iterator tmpi = i;
	    ++i;
	    ++ri;
	    a = i;
	    reading->erase(tmpi);
	    doInsertion = true; 
	  }
	  else {
	    ++i;
	  }
	}
      // Only carry out the insertion if anything was removed
      if(doInsertion) reading->insert(a, insertion->begin(), insertion->end());
      return doInsertion;
    }
  
  qerr << "Correction::correct : We should never be here!\n";
  return false;
}

class TargetMatchesNot
{
  Set * target;
public:
  TargetMatchesNot(Set * t) : target(t) { }
  bool operator() (Combined* reading) 
  {
    //    cout << *target << "->matches(";
    //    reading->put(cout);
    //    cout << ") = " << target->matches(reading) << "\n";
    return !target->matches(reading);
  }
};

/**
 * Try to apply a Disambiguation rule to a reading of a Cohort.
 */
bool Constraint::disambiguate(vector<Cohort*>* cohorts, int pos, list<Combined*>::iterator ri) const
{
  Cohort * cohort = (*cohorts)[pos];
  Combined * reading = *ri;
  
  // Does the wordform match?
  if(! (wordform==*empty_string || wordform==cohort->getWordform())) return false;
  
  // Does the target match?
  if(! target->matches(reading) ) return false;
  
  //cout << toString() << " with target: "<<cohort->toString()<<"\n";
  
  // Do all context tests match?
  list<Context*>::iterator a = tests->begin();
  for(list<Context*>::iterator
	i = a,
	z = tests->end();
      i!=z; ++i)
    {
      if(! (*i)->matchCohort(cohorts,pos))
	{
	  // Move non-matching tests to the front, 
	  // assuming that they will fail more often than other tests.
	  // Thus most often the _first_ constraint tested will
	  // be the one that fails (saving some time)
	  // [ this increased performance by appx 10-15% ]
	  iter_swap(a,i);
	  
	  return false;
	}
    }
  
  // Carry out the operation
  list<Combined*>* readings = cohort->getReadings();
  bool doRemoveThis = false;
  bool operationApplied = false;
  
  switch(op)
    {
    case op_remove:
      
#ifndef BINARY
      // Debug mode
      if(Debug::instance()->debug) {
	if((*ri)->has(CGCompiler::correctTag())) 
	  {
	    cout << "WARNING: ";
	    (*ri)->put(cout);
	    cout << (Debug::instance()->debug_cg2 ?
		     " may be removed by " : 
		     " removed by ");
	    cout << (*this) << "\n";
	  }
	if(Debug::instance()->debug_cg2) return false;
      }
#endif
      
      // Test whether this rule targets all other readings.
      // If so, none of them are removed.
      for(list<Combined*>::iterator
	    i = readings->begin(),
	    z = readings->end();
	  i!=z; ++i)
	{
	  if(i == ri) continue;
	  if(! target->matches(*i) ) {
	    doRemoveThis = true;
	    break;
	  }
	}
      if(!doRemoveThis) return false;
      
      // Do the removal.
      delete *ri;
      readings->erase(ri);
      if(Debug::instance()->minimal)
	{
	  Combined * wfr = cohort->getWordformReading();
	  if(wfr->size()==1)
	    {
	      string * tag = new string("S:"+(*(itos(line))));
	      wfr->push_back(TagRegister::instance()->add(*tag));
	    }
	  else
	    {
	      Combined::iterator i = wfr->begin();
	      ++i;
	      // BUG? This looks dangerous.
	      // What happens if 2 cohorts have identical trace tags?
	      string s = TagRegister::instance()->get(*i);
	      s.append(", "+(*(itos(line))));
	      wfr->erase(i);
	      wfr->push_back(TagRegister::instance()->add(s));
	    }
	}
      return true;
      
    case op_select:

#ifndef BINARY
      // Debug mode
      if(Debug::instance()->debug) 
	{
	  for(list<Combined*>::iterator
		i = readings->begin(),
		z = readings->end();
	      i!=z; ++i)
	    {
	      if(i == ri || target->matches(*i))
		continue;
	      else {
		if((*i)->has(CGCompiler::correctTag())) {
		  cout << "WARNING: ";
		  (*i)->put(cout);
		  cout << (Debug::instance()->debug_cg2 ?
			   " may be removed by " : 
			   " removed by ");
		  cout << (*this) << "\n";
		}
	      }
	    }
	  if(Debug::instance()->debug_cg2) return false;
	}
#endif
      
      // Test whether this rule selects other readings.
      // All selected readings must be retained.
      
      // remove all readings before the selected one. 
      // BUG: this will actually remove all selectable readings before this one
      //if(readings->begin() != ri) operationApplied = true;
      //readings->erase(readings->begin(), ri);
      
      // remove all non-matching readings
      list<Combined*>::iterator
	ti = readings->begin(),
	tz = readings->end();
      while( ti!=tz )
	{
	  list<Combined*>::iterator i = ti;
	  ++ti;
	  
	  // remove reading if it doesn't match
	  if(! target->matches(*i) )
	    {
	      delete *i;
	      readings->erase(i);
	      operationApplied = true;
	    }
	}
      
      if(operationApplied && Debug::instance()->minimal)
	{
	  Combined * wfr = cohort->getWordformReading();
	  if(wfr->size()==1)
	    {
	      // BUG: Minor memory leak possible?
	      string * tag = new string("S:"+(*(itos(line))));
	      wfr->push_back(TagRegister::instance()->add(*tag));
	    }
	  else
	    {
	      Combined::iterator i = wfr->begin();
	      ++i;
	      // BUG? This looks dangerous.
	      // What happens if 2 cohorts have identical trace tags?
	      string s = TagRegister::instance()->get(*i);
	      s.append(", "+(*(itos(line))));
	      wfr->erase(i);
	      wfr->push_back(TagRegister::instance()->add(s));
	    }
	}
      
      return operationApplied;
    }
  
  qerr << "Constraint::disambiguate : We should never be here!\n";
  return false;
}

/**
 * Try to apply a mapping rule to a Cohort.
 */
bool Mapping::map(vector<Cohort*>* cohorts, int pos, Combined * reading) const
{
  //Cohort * cohort = (*cohorts)[pos];
  
  // Does the target match?
  if(! target->matches(reading) ) return false;
  
  // Do all context tests match?
  list<Context*>::iterator a = tests->begin();
  for(list<Context*>::iterator
	i = a,
	z = tests->end();
      i!=z; ++i)
    {
      if(! (*i)->matchCohort(cohorts,pos)) 
	{
	  // Move non-matching tests to the front, 
	  // assuming that they fail more often than other tests.
	  // Thus most often the _first_ constraint tested will
	  // be the one that fails (saving some time)
	  iter_swap(a,i);
	  
	  return false;
	}
    }
  
#ifndef BINARY
  // Debug mode
  if(Debug::instance()->debug) {
    //Combined * correct = new Combined();
    //string * ctag = new string("<Correct!>");
    //correct->push_back(TagRegister::instance()->add(*ctag));
    if(reading->has(CGCompiler::correctTag())) 
      {
	cout << "WARNING: ";
	reading->put(cout);
	cout << (Debug::instance()->debug_cg2 ?
		 " may be modified by " : 
		 " modified by ");
	cout << (*this) << "\n";
      }
    if(Debug::instance()->debug_cg2) return false;
  }
#endif
  

  // Carry out the operation.
  switch(op)
    {
    case op_replace:
      {
	// clear everything except first tag.
	Combined::iterator si = reading->begin();
	++si;
	reading->erase(si,reading->end()); 
      }
    case op_add:
    case op_map:
      if(Debug::instance()->minimal)
	{
	  string * tag = new string("S:"+(*(itos(line))));
	  reading->push_back(TagRegister::instance()->add(*tag));
	}
      // add insertion tags to the reading.
      reading->insert(reading->end(), insertion->begin(), insertion->end());
      return true;
    }
  
  qerr << "Mapping::map : We should never be here!\n";
  return false;
}

/**
 * Test whether a context test matches the specified Cohort.
 * 
 * NOTE:
 * 
 * A: In negated contexts, such as (NOT context0 LINK context1 LINK context2), 
 *    the negation is applied last. (see 2.4.5. page 33: "Here, the negation is applied last").
 *    The above context test is interpreted as 
 *         ! (context0 && (context1 && context2))
 * 
 * B: However, in negated LINKed contexts such as 
 *         (context0 LINK NOT context1 LINK context2),
 *    the negation is applied only to context1, not to "context1 LINK context2"
 *    I.e.: the LINK to context2 will only be tested if context0 matches and 
 *    the linked context1 does not.
 *    The above context test is interpreted as 
 *         context0 && ( (!context1) && context2)
 *
 * C: Another example:
 *         (NOT context0 LINK NOT context1 LINK context2)
 *    The above context test is interpreted as 
 *         ! (context0 && ( (!context1) && context2))
 * 
 * It's a bit odd, but this seems to be the way it works in CG-2. 
 * The asLink parameter controls whether negation is interpreted as 
 * case A (asLink==false) or case B (asLink==true).
 */
bool Context::matchCohort(vector<Cohort*>* cohorts, unsigned int pos, bool asLink=false) const
{
  int offset = position->getOffset();
  bool careful = position->isCareful();
  
  // Decide position in input
  if(position->isAbsolute()) // absolute
    pos = offset < 0 ? cohorts->size()-offset-1 : offset;
  else // relative
    pos += position->getOffset();
  
  // Position out of bounds?
  if(pos < 0 || pos >= cohorts->size()) return false ^ negative;
  
  // Determine scan direction
  int direction = offset<0 ? -1 : 1;
  
  // Do the test
  switch(position->getScantype())
    {
    case scan_none:
      // No scan. Just test the cohort at the position.
      {
	Cohort * cohort = (*cohorts)[pos];
	
	bool match =
	  testset->matches(cohort->getWordformReading()) ||
	  testset->matches(cohort, careful);
	match =  asLink
	  ? negative ^ match
	  : match;
	
	// test the link if there was a match
	if(match && link!=0) match = link->matchCohort(cohorts,pos,true);
	
	return asLink ? match : negative ^ match;
      }
      
    case scan_asterisk:
      // Simple scan. Stops at the first match or barrier.
      for(unsigned int i = pos;
	  (i>=0 && i<cohorts->size()); 
	  i+=direction
	  )
	{
	  Cohort * cohort = (*cohorts)[i];
	  
	  bool match = 
	    testset->matches(cohort->getWordformReading()) ||
	    testset->matches(cohort, careful);
	  bool barr  = barrier==0 ? false : barrier->matches(cohort) ;
	  
	  //if(barrier && !barr) cout << "BARRIER: " << *barrier << " !matches " << *cohort << " in " << *this << "\n";
	  
	  // barrier reached without a match -> end scan
	  if( barr && !match) break;
	  // no match yet -> continue scan
	  if(!match) continue;
	  // test the link
	  if(match && link!=0) match = link->matchCohort(cohorts,i,true);
	  
	  // end scan at first match
	  return match ^ negative;
	}
      return false ^ negative;
      
    case scan_doubleasterisk:
      // Continuing scan. Doesn't give up until the end of input (or when hitting a barrier).
      for(unsigned int i = pos;
	  i>=0 && i<cohorts->size(); 
	  i+=direction
	  )
	{
	  Cohort * cohort = (*cohorts)[i];
	  
	  bool match = 
	    testset->matches(cohort->getWordformReading()) ||
	    testset->matches(cohort, careful);
	  bool barr  = barrier==0 ? false : barrier->matches(cohort) ;
	  
	  // barrier reached without a match -> end scan
	  if(barr && !match) break;
	  // no match yet -> continue scan
	  if(!match) continue; 
	  // test the link
	  if(match && link!=0) match = link->matchCohort(cohorts,i,true);
	  
	  if(match) return true ^ negative;
	  if(barr) return false ^ negative;
	}
      return false ^ negative;
    }
  
  qerr << "Context::matchCohort : We should never be here!";
  return false;
}


/**
 * Determine whether this combined tag has all the tags
 * of the specified combined tag. The order of the tags in the
 * two combined tags must be the same.
 * E.g. (A B C D E) has (B D E) but not (D E F) and not (B A).
 */
template<class T> 
bool taglist<T>::has(taglist<T> * tags) const
{
  for(taglist<T>::const_iterator
	ri = begin(),
	rz = end(),
	ti = tags->begin(),
	tz = tags->end();
      ri!=rz; ++ri)
    {
      if((*ri)==(*ti)) ++ti;
      if(ti==tz) return true; // all tags were matched.
    }
  return false;
}

TagRegister * TagRegister::_instance = 0;
TagRegister * TagRegister::instance()
{
  if (_instance == 0)
    _instance = new TagRegister();
  return _instance;
}



//
// Sets.
//

//
// BUG:
// 
// Of the *::matches(Cohort, bool careful) methods, only 
// SetReference::matches(Cohort, bool careful) is ever invoked.
// 
// All other *::matches(Cohort, bool careful) should not be trusted
// and are unlikely to work correctly with careful matching.
// 

bool SetReference::targets(Combined * tag) const
{
  return decl->targets(tag);
}
bool SetReference::matches(Combined * tag) const
{
  return decl->matches(tag);
}
/**
 * SetReference::matches(...) delegates the membership tests for each
 * reading of the cohort to the matches(Combined) method and handles 
 * carefulness.
 */
bool SetReference::matches(Cohort * cohort, bool careful=false) const
{
  if(matches(cohort->getWordformReading())) return true;
  if(cohort->getReadings()->size() == 0) return false;
  for(list<Combined*>::iterator
	i = cohort->getReadings()->begin(),
	z = cohort->getReadings()->end();
      i!=z; ++i)
    {
      if(matches(*i))
	{ if(!careful) return true; }
      else 
	{ if(careful) return false; }
    }
  return careful;
}

bool SetDeclaration::targets(Combined * tag) const
{
  return expr->targets(tag);
}
bool SetDeclaration::matches(Combined * tag) const
{
  return expr->matches(tag);
}
bool SetDeclaration::matches(Cohort * cohort, bool careful=false) const
{
  cerr << "WARNING: Internal error: SetDeclaration::matches(Cohort, bool) invoked.\n";
  if(matches(cohort->getWordformReading())) return true;
  if(cohort->getReadings()->size() == 0) return false;
  for(list<Combined*>::iterator
	i = cohort->getReadings()->begin(),
	z = cohort->getReadings()->end();
      i!=z; ++i)
    {
      if(matches(*i))
	{ if(!careful) return true; }
      else 
	{ if(careful) return false; }
    }
  return false;
}

bool ListDeclaration::targets(Combined * tag) const
{
  for(list<Combined*>::iterator
	i = combinedTags->begin(),
	z = combinedTags->end();
      i!=z; ++i)
    {
      if((*i)->has(tag)) return true;
    }
  return false;
}
bool ListDeclaration::matches(Combined * reading) const
{
  for(list<Combined*>::iterator
	i = combinedTags->begin(),
	z = combinedTags->end();
      i!=z; ++i)
    {
      if(reading->has(*i)) return true;
    }
  return false;
}
bool ListDeclaration::matches(Cohort * cohort, bool careful=false) const
{
  cerr << "WARNING: Internal error: ListDeclaration::matches(Cohort, bool) invoked.\n";
  if(matches(cohort->getWordformReading())) return true;
  if(cohort->getReadings()->size() == 0) return false;
  for(list<Combined*>::iterator
	i = cohort->getReadings()->begin(),
	z = cohort->getReadings()->end();
      i!=z; ++i)
    {
      if(matches(*i)) 
	{ if(!careful) return true; }
      else 
	{ if(careful) return false; }
    }
  
  return careful;
}

bool SetUnion::targets(Combined * reading) const
{
  if(set1->targets(reading)) return true;
  if(set2->targets(reading)) return true;
  return false;
}
bool SetUnion::matches(Combined * reading) const
{
/*
  cout << *this << " on " ;
  reading->put(cout);
  cout <<" : " << *set1 << " " << set1->matches(reading) << "\n";
  cout << ": " << *set2 << " " << set2->matches(reading) << "\n";
*/
  if(set1->matches(reading)) return true;
  if(set2->matches(reading)) return true;
  return false;
}
bool SetUnion::matches(Cohort * cohort, bool careful=false) const
{
  cerr << "WARNING: Internal error: SetUnion::matches(Cohort, bool) invoked.\n";
  if(matches(cohort->getWordformReading())) return true;
  if(cohort->getReadings()->size() == 0) return false;
  if(set1->matches(cohort)) return true;
  if(set2->matches(cohort)) return true;
  return false;
}

bool SetConcatenation::targets(Combined * reading) const
{
  if(!set1->targets(reading)) return false;
  if(set2->targets(reading)) return true;
  return false;
}
bool SetConcatenation::matches(Combined * reading) const
{
  /*
  cout << *this << " on " ;
  reading->put(cout);
  cout <<" : " << *set1 << " " << set1->matches(reading) << "\n";
  cout << ": " << *set2 << " " << set2->matches(reading) << "\n";
  */
  if(!set1->matches(reading)) return false;
  if(set2->matches(reading)) return true;
  return false;
}
bool SetConcatenation::matches(Cohort * cohort, bool careful=false) const
{
  cerr << "WARNING: Internal error: SetConcatenation::matches(Cohort, bool) invoked.\n";
  if(matches(cohort->getWordformReading())) return true;
  if(cohort->getReadings()->size() == 0) return false;
  for(list<Combined*>::iterator
	i = cohort->getReadings()->begin(),
	z = cohort->getReadings()->end();
      i!=z; ++i)
    {
      if(matches(*i)) 
	{ if(!careful) return true; }
      else 
	{ if(careful) return false; }
    }
  return careful;
}

bool SetDifference::targets(Combined * reading) const
{
  if(!set1->targets(reading)) return false;
  if(!set2->targets(reading)) return true;
  return false;
}
bool SetDifference::matches(Combined * reading) const
{
  /*
  cout << *this << " on " ;
  reading->put(cout);
  cout <<" : " << *set1 << " " << set1->matches(reading) << "\n";
  cout << ": " << *set2 << " " << set2->matches(reading) << "\n";
  */
  if(!set1->matches(reading)) return false;
  if(!set2->matches(reading)) return true;
  return false;
}
bool SetDifference::matches(Cohort * cohort, bool careful=false) const
{
  cerr << "WARNING: Internal error: SetDifference::matches(Cohort, bool) invoked.\n";
  if(matches(cohort->getWordformReading())) return true;
  if(cohort->getReadings()->size() == 0) return false;
  for(list<Combined*>::iterator
	i = cohort->getReadings()->begin(),
	z = cohort->getReadings()->end();
      i!=z; ++i)
    {
      if(matches(*i)) 
	{ if(!careful) return true; }
      else 
	{ if(careful) return false; }
    }
  return careful;
}

//
// Stream output.
//

string * ltltlt = new string("<<<");
int endMarker = TagRegister::instance()->add(*ltltlt);

template<class T> 
ostream& taglist<T>::put(ostream& os, bool noParen=true, hash_map<T,string>* tagmap=0) const
{
  if(tagmap==0) tagmap = TagRegister::instance()->getReverseMap();
  
  if(!noParen) os << "(";
  const_iterator
    i=begin(),
    z=end();
  if((*i)==endMarker) ++i;
  for(; i!=z; ++i)
    {
      if(i!=begin()) os << " ";
      if(tagmap)
	{
	  hash_map<T,string>::iterator p = tagmap->find(*i);
	  if(p==tagmap->end()) os << '?' << (*i) << '?';
	  else os << p->second;
	}
      else
	{
	  os << (*i);
	}
    }
  if(!noParen) os << ")";
  return os;
}



ostream& operator<< (ostream& os, const Set& s)
{
  return s.put(os);
}
ostream& SetReference::put(ostream& os) const
{
  if(setname == "") return os << (*decl);
  return os << setname;
}
ostream& SetDeclaration::put(ostream& os) const
{
  return os << "SET" << setname << " = " << (*expr);
}
ostream& ListDeclaration::put(ostream& os) const
{
  if(listname!=*empty_string) os << "LIST " << listname << " = ";
  bool space = false;
  for(list<Combined*>::iterator
	i = combinedTags->begin(),
	z = combinedTags->end();
      i!=z; ++i)
    {
      if(space) os << " ";
      else space=true;
      os << '(';
      (*i)->put(os);
      os << ')';
    }
  return os;
}
ostream& SetUnion::put(ostream& os) const
{
  return os << (*set1) << " OR " << (*set2);
}
ostream& SetConcatenation::put(ostream& os) const
{
  return os << (*set1) << " + " << (*set2);
}
ostream& SetDifference::put(ostream& os) const
{
  return os << (*set1) << " - " << (*set2);
}



ostream& operator<< (ostream& os, const Cohort& c)
{
  c.wf_reading->put(os);
  os << '\n';
  if(c.readings==0) return os << "\tnull";
  for(list<Combined*>::iterator 
	i=c.readings->begin(),
	z=c.readings->end();
      i!=z; ++i )
    {
      if(*((*i)->begin())==endMarker) continue;
      os << '\t';
      (*i)->put(os);
      os << '\n';
    }
  return os;
}
ostream& operator<< (ostream& os, const Position& p)
{
  if(p.absolute) os << "@";
  if(p.scantype == scan_asterisk) os << "*";
  if(p.scantype == scan_doubleasterisk) os << "**";
  os << p.offset;
  if(p.careful) os << "C";
  return os;
}
ostream& operator<< (ostream& os, const Context& c)
{
  os << (c.negative? "NOT " : "");
  if(c.position==0) os << "null"; else os << (*(c.position)) << " ";
  if(c.testset ==0) os << "null"; else os << (*(c.testset));
  if(c.barrier !=0) os << " BARRIER " << (*(c.barrier));
  if(c.link    !=0) os << " LINK " << (*(c.link));
  return os;
}
ostream& operator<< (ostream& os, const Mapping& m)
{
  if(m.op==op_map) os << "MAP";
  if(m.op==op_add) os << "ADD";
  if(m.op==op_replace) os << "REPLACE";
  os << " (";
  m.insertion->put(os);
  os << ')';
  if(m.target==0) os << " null"; else os << " TARGET " << (*(m.target));
  if(m.tests==0) 
    { 
      os << " null "; 
    }
  else if(m.tests->size()==0)
    {
      os << " ";
    }
  else
    {
      os << " IF ";
      for(list<Context*>::iterator
	    i = m.tests->begin(), 
	    z = m.tests->end();
	  i!=z; ++i)
	{
	  if((*i)==0) os << "null ";
	  else os << '(' << (**i) << ") ";
	}
    }
  os <<  ";";
  if(m.line!=0) os << " # line " << m.line;
  return os;
}
ostream& operator<< (ostream& os, const Constraint& c)
{
  if(c.wordform!="") os << c.wordform << " ";
  if(c.op==op_remove) os << "REMOVE";
  if(c.op==op_select) os << "SELECT";
  if(c.target==0) os << " null"; else os << " TARGET " << (*(c.target)) << ' ';
  if(c.tests==0)
    {
      os << "null ";
    }
  else if(! c.tests->size()==0)
    {
      os << "IF ";
      for(list<Context*>::iterator
	    i=c.tests->begin(), 
	    z=c.tests->end();
	  i!=z; ++i)
	{
	  if((*i)==0) os << "null ";
	  else os << "(" << (**i) << ") ";
	}
    }
  os << ";";
  if(c.line!=0) os << " # line " << c.line;
  return os;
}
ostream& operator<< (ostream& os, const Correction& c)
{
  if(c.wordform!="") os << c.wordform << " ";
  if(c.op==op_substitute) os << "SUBSTITUTE";
  if(c.op==op_append) os << "APPEND";
  if(c.removal!=0)
    {
      os << " (";
      c.removal->put(os);
      os << ')';
    }
  if(c.insertion!=0)
    {
      os << " (";
      c.insertion->put(os);
      os << ')';
    }
  if(c.target==0) os << " null "; else os << " TARGET " << (*(c.target)) << ' ';
  if(c.tests==0)
    {
      os << "null ";
    }
  else if(! c.tests->size()==0)
    {
      os << "IF ";
      for(list<Context*>::iterator
	    i=c.tests->begin(), 
	    z=c.tests->end();
	  i!=z; ++i)
	{
	  if((*i)==0) os << "null ";
	  else os << "(" << (**i) << ") ";
	}
    }
  os << ";";
  if(c.line!=0) os << " # line " << c.line;
  return os;
}

//
// Constructors
//



//
// Destructors
//
Cohort::~Cohort()
{
  for(list<Combined*>::iterator
	i=readings->begin(), z=readings->end();
      i!=z; ++i)
    {
      delete *i;
    }
  delete readings;
  delete wf_reading;
}
