%{
/* 
 * The contents of this file are subject to the GrammarSoft Public
 * License Version 1.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.grammarsoft.com/GSPL
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is the VISL CG Parser.
 * 
 * The Initial Developer of the Original Code is Martin Carlsen, 
 * GrammarSoft, and SDU Odense University.  
 * Portions created by Martin Carlsen, Grammarsoft, and SDU Odense 
 * University are Copyright (C) 1999-2000 Martin Carlsen, GrammarSoft, 
 * and SDU Odense University.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * Alternatively, the contents of this file may be used under the
 * terms of the GNU General Public License Version 2 or later (the
 * "GPL"), in which case the provisions of the GPL are applicable 
 * instead of those above.  If you wish to allow use of your 
 * version of this file only under the terms of the GPL and not to
 * allow others to use your version of this file under the GSPL,
 * indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by
 * the GPL.  If you do not delete the provisions above, a recipient
 * may use your version of this file under either the GSPL or the
 * GPL.
 */
  /*
   * This file contains the Bison rules for the parser 
   * for the Constraint Grammar rules.
   *
   * To build a CG compiler, call buildCompiler with an input stream 
   * reading from the rule definitions as argument.
   */
#include <map>
#include <string>
#include <list>
#include <set>  
#include <iostream>
  
#include "cg.data_structures.hh"
#include "CGCompiler.hh"
#include "Options.hh"
#include "Debug.hh"
#include "Generator.hh"
  
#undef yyFlexLexer
#define yyFlexLexer cgrulesFlexLexer
#include <FlexLexer.h>
  
  FlexLexer * lexer = new cgrulesFlexLexer();
  
#define yyparse yyparse
#define yylex lexer->yylex
#define yyerror yyerror
#define yydebug lexer->yydebug
  
  extern int numlines; // from scanner.l
  extern int balance; // from scanner.l
  extern string * startTag; // from cohorts.cc
  extern string * endTag; // from cohorts.cc
  TagRegister * tag_register = TagRegister::instance();
  Debug * debug = Debug::instance();
  
  taglist<int> * delimiterlist;
  taglist<int> * preferredlist;
  map<string,Set*> * sets;
  list<list<Constraint*>*> * constraintsections;
  list<Correction*> * corrections;
  list<Mapping*> * mappings;
  list<Set*>* setList;
  
  Generator * generator = 0;
  
  int main(int argc, char * argv[]);
  
  int yyerror(char *s)
    {
      cerr << "vislcg: *** syntax error in line "<< numlines ;
      debug->check_only = true;
      debug->verbose = false;
      return 1;
    }
  %}

%union
{
  Set * set;
  Position * position;
  Context * context;
  Correction * correction;
  Mapping * mapping;
  Constraint * constraint;
  list<Constraint*> * constraints;
  string * tag;
  Combined * combined;
  Combined * taglist;
  list<Combined*> * tags;
  list<Context*> * tests;
}

%token        NUM
%token        <tag> ID
%token        delimiters
%token        preferred
%token        END
%token        LIST
%token        SET
%token        ADD
%token        MAP
%token        REPLACE
%token        AND
%left         OR
%token        NOT
%token        doubleasterisk
%token        REMOVE
%token        ASSIGNOP
%token        LINK
%token        BARRIER
%token        posnum
%token        SELECT
%token        IF
%token        TARGET
%token        SETS
%token        CORRECTIONS
%token        APPEND
%token        SUBSTITUTE
%token        MAPPINGS
%token        CONSTRAINTS 
%token        <position> POSITION

%type <tag> wordform
%type <tags> tags
%type <taglist> tag_items
%type <combined> combined
%type <mapping> mapping
%type <set> set_expr set_term set_definition barrier target
%type <correction> correction;
%type <constraint> constraint
%type <constraints> constraints
%type <context> context link and_context
%type <tests> contextlist
%%

/**
 * Rule file (2.1)
 */
program: delimiters_section preferred_section sections optional_end
{
  rout << "RECOGNIZED RULE FILE\n";
  
if(!debug->quiet)
  {
    
    dout(debug->verbose) << "DELIMITERS = ";
    if(debug->verbose) delimiterlist->put(cout); 
    dout(debug->verbose) << " ;";
    
    dout(debug->verbose) << "\n\nPREFERRED-TARGETS = ";
    if(debug->verbose) preferredlist->put(cout);
    dout(debug->verbose) << " ;";
    
    dout(debug->verbose) << "\n\nSETS\n\n";
    for(map<string,Set*>::iterator 
	  i = sets->begin(),
	  z = sets->end();
	i!=z; ++i)
      {
	dout(debug->verbose) << (*((*i).second)) << "\n";
      }
    
    dout(debug->verbose) << "\nMAPPINGS\n\n";
    for(list<Mapping*>::iterator 
	  i = mappings->begin(),
	  z = mappings->end();
	i!=z; ++i)
      {
	dout(debug->verbose) << (**i) << "\n";
      }
    
    int count = 0;
    for(list<list<Constraint*>*>::iterator 
	  i2 = constraintsections->begin(),
	  z2 = constraintsections->end();
	i2!=z2; ++i2)
      {
	dout(debug->verbose) << "\nCONSTRAINTS\n\n";
	count += (*i2)->size();
	for(list<Constraint*>::iterator 
	      i = (*i2)->begin(),
	      z = (*i2)->end();
	    i!=z; ++i)
	  {
	    dout(debug->verbose) << (**i) << "\n";
	  }
      }
    
    dout(debug->verbose) << "\nEND\n";
    
    qerr << "\n";
    qerr << sets->size() << " sets (using "<<tag_register->getReverseMap()->size()<<" different tags)\n";
    qerr << mappings->size() << " mappings\n";
    qerr << constraintsections->size() << " constraint sections\n";
    qerr << count << " constraint rules\n";
  }
}
;

/**
 * Optional END keyword
 */
optional_end: END
|
;


//
// Sections
//


/**
 * Delimiter section (2.1.1)
 */
delimiters_section:   delimiters ASSIGNOP tag_items ';' 
{
  rout << "delimiters = tag_items ;\n";
  delimiterlist = $3; 
  delimiterlist->reverse(); 
}
;

/**
 * Preferred-targets section (2.1.2)
 */
preferred_section:   preferred ASSIGNOP tag_items ';'
{
  rout << "preferred-targets = tag_items ;\n";
  preferredlist = $3; 
  preferredlist->reverse(); 
}
|
{
  preferredlist = new Combined();
}
;

/**
 * The SETS, CORRECTIONS, MAPPINGS, and CONSTRAINTS sections:
 */
sections: section sections
| section
;
section: SETS sets_and_lists
{
  rout << "SETS\n";
}
| CORRECTIONS corrections
{
  rout << "CORRECTIONS\n";
}
| CONSTRAINTS constraints
{
  rout << "CONSTRAINTS\n";
  constraintsections->push_back($2);
}
| MAPPINGS mappings
{
  rout << "MAPPINGS\n";
}
;

/**
 * List of zero or more correction specifications.
 */
corrections: correction corrections
{
  corrections->push_front($1) ;
}
|
;

/**
 * List of zero or more mapping specifications.
 */
mappings: mapping mappings 
{ 
  mappings->push_front($1) ;
}
|
;

/**
 * List of zero or more constraint specifications.
 */
constraints: constraint constraints
{
  rout << (*($1)) << "\n";
  
  $$ = $2;
  $$->push_front($1);
}
|
{
  $$ = new list<Constraint*>();
}
;

/**
 * List of zero or more SET and LIST definitions
 */
sets_and_lists: named_set_or_list sets_and_lists
|
;


//
// Sets and set expressions.
//


/**
 * Named SET or LIST definition.
 */
named_set_or_list: LIST ID ASSIGNOP tags ';' 
{
  rout << "LIST " << *($2) << " = tags ;\n";
  if ((*sets)[*($2)]) 
    { 
      qerr << "vislcg: *** error: in line " << numlines << ": " << *($2) << " already defined.\n"; 
      debug->verbose = false;
      debug->check_only = true;
    }
  Set * s = new ListDeclaration($4,*($2));
  (*sets)[*($2)] = s;
  rout << (*s) << "\n";
  if(generator) generator->addSet(s);
}
| SET ID ASSIGNOP set_definition ';' 
{
  rout << "SET ID ASSIGNOP set_definition ;\n";
  if ((*sets)[*($2)]) { 
    qerr << "vislcg: *** error: in line " << numlines << ": " << *($2) << " already defined.\n";
    debug->verbose = false;
    debug->check_only = true;
  }
  Set * s = new SetDeclaration($4,*($2));
  (*sets)[*($2)] = s;
  rout << (*s) << "\n";
  if(generator) generator->addSet(s);
}
;

/**
 * Anonymous set definition.
 */
set_definition: set_expr
{
  rout << "set_expr\n";
  $$ = $1;
}
| set_definition OR set_expr
{
  rout << "set_definition OR set_expr\n";
  $$ = new SetUnion($1, $3);
}
/*
| set_definition AND set_expr
{
  rout << "set_definition AND set_expr\n";
  qerr << "vislcg: *** error: in line " << numlines << ": ambiguity class (AND) not supported, use LINK 0 instead.\n";
  debug->verbose = false;
  debug->check_only = true;
  $$ = new SetConcatenation($1, $3);
}
*/
;

/**
 * An expression that operates on sets.
 */
set_expr: set_expr '+' set_term
{
  rout << "set_expr + set_term\n";
  $$ = new SetConcatenation($1, $3);
}
| set_expr '-' set_term
{
  rout << "set_expr - set_term\n";
  $$ = new SetDifference($1, $3);
}
| set_term
{
  rout << "set_term\n";
  $$ = $1;
}
;

/**
 * A reference to a named set or a set of a single combined tag.
 */
set_term: combined
{
  rout << "combined\n";
  list<Combined*>* single = new list<Combined*>();
  single->push_back($1);
  $$ = new ListDeclaration(single); 
}
| ID
{
  rout << "ID("<<*($1)<<"\n";
  if ( sets->find(*($1)) == sets->end() )
    {
      qerr << "vislcg: *** error: in line "<<numlines<<": undefined set or list: "<<(*$1) << " (spelling error or missing parentheses?)\n";
      debug->verbose = false;
      debug->check_only = true;
    }
  $$ = new SetReference((*sets)[*($1)], *($1));
}
;


//
// Tags and lists of tags.
//


/**
 * Combined tag, such as (N NOM SG)
 */
combined: '(' tag_items ')'
{
  rout << "( tag_items )\n";
  $$ = $2;
  $$->reverse();
}
;

/**
 * A list of blank-separated tag items
 */
tag_items:   ID
{
  rout << "tag(" << *($1) << ")\n";
  
  $$ = new Combined(); 
  int t = tag_register->add(*($1)) ;
  $$->push_back(t);
}
| ID tag_items
{
  rout << "tag(" << *($1) << ")\n";
  $$ = $2; 
  $$->push_back(tag_register->add(*($1)));
}
;

/**
 * List of (mixed combined and single) tags.
 */
tags: combined tags
{
  rout << "combined tags\n";
  $$ = $2;
  $$->push_back($1);
}
| combined
{
  rout << "combined\n";
  $$ = new list<Combined*>();
  $$->push_back($1);
}
| ID tags
{
  rout << "ID("<<*($1)<<") tags\n";
  tag_register->add(*($1));
  Combined * single = new Combined();
  single->push_back(tag_register->add(*($1)));
  $$ = $2;
  $$->push_back(single);
}
| ID
{
  rout << "ID("<<*($1)<<")\n";
  Combined * single = new Combined();
  single->push_back(tag_register->add(*($1)));
  $$ = new list<Combined*>();
  $$->push_back(single);
}
;


//
// Rules for mapping and disambuation.
//

/**
 * Correction rule. (SUBSTITUTE, APPEND)
 */
correction: wordform SUBSTITUTE combined combined target contextlist ';'
{
  rout << "wordform SUBSTITUTE combined combined target contextlist\n";
  $$ = new Correction( *($1), op_substitute, $3, $4, new SetReference($5), $6, numlines);
}
| wordform APPEND combined contextlist ';'
{
  rout << "wordform APPEND combined contextlist\n";
  $$ = new Correction( *($1), op_append, 0, $3, 0, $4, numlines);
}
;

/**
 * Mapping rule. (MAP, ADD, REPLACE)
 */
mapping: MAP combined target contextlist ';' 
{ 
  rout << "MAP combined set_definition contextlist\n";
  $$ = new Mapping(op_map, $2, new SetReference($3), $4, numlines);
}
| ADD combined target contextlist ';' 
{
  rout << "ADD combined set_definition contextlist\n";
  $$ = new Mapping(op_add, $2, new SetReference($3), $4, numlines);
}
| REPLACE combined target contextlist ';' 
{
  rout << "REPLACE combined set_definition contextlist\n";
  $$ = new Mapping(op_replace, $2, new SetReference($3), $4, numlines);
}
;

/**
 * Disambiguation rule. (REMOVE, SELECT)
 */
constraint: wordform REMOVE target contextlist ';' 
{
  $$ = new Constraint ( *($1), op_remove, new SetReference($3), $4, numlines);
}
| wordform SELECT target contextlist ';' 
{
  $$ = new Constraint (*($1), op_select, new SetReference($3), $4, numlines);
}
;

/**
 * Target specification, with optional TARGET keyword.
 */
target: TARGET set_definition
{
  $$ = $2;
}
| set_definition
{
  $$ = $1;
}
;

/**
 * Optional wordform specification
 */
wordform: ID
{
  $$ =  $1;
}
| 
{ 
  $$ = new string(); 
}
;

/**
 * List of zero or more contextual tests, with optional IF keyword
 */
contextlist: 
{
  rout << "(no context)\n";
  $$ = new list<Context*>();
}
| IF contextlist
{
  rout << "IF contextlist\n";
  $$ = $2;
}
| '(' context ')' contextlist 
{
  rout << "( context ) contextlist\n";
  $$ = $4;
  $$->push_back($2);
}
;

/**
 * Contextual test
 */
context: POSITION set_definition barrier link
{
  rout << "POSITION set_definition barrier("<<$3<<") link\n";
  $$ = new Context(false, $1, new SetReference($2), $3, $4);
  rout << numlines << (*($$)) << "\n";
}
| NOT POSITION set_definition barrier link
{
  rout << "NOT POSITION set_definition barrier link\n";
  $$ = new Context(true, $2, new SetReference($3), $4, $5);
  rout << numlines << (*($$)) << "\n";
}
;

/**
 * Optional BARRIER specification.
 */
barrier: BARRIER set_definition 
{
  rout << "BARRIER set_definition\n";
  $$ = new SetReference($2); 
  rout << (*($$)) << "\n";
}
|
{
  rout << "(no barrier)\n";
  Set * nobarrier = 0; 
  $$ = nobarrier; 
}
;

/**
 * AND context (AND is interpreted as LINK 0)
 */
and_context: set_definition barrier link
{
  rout << "set_definition barrier("<<$3<<") link\n";
  $$ = new Context(false, new Position, new SetReference($1), $2, $3);
  rout << numlines << (*($$)) << "\n";
}
;

/**
 * Optional LINK specification.
 */
link: LINK context
{
  rout << "LINK context\n";
  $$ = $2; 
}
| AND and_context
{
  rout << "AND and_context\n";
  $$ = $2;
}
|
{
  rout << "(no link)\n";
  Context * nolink = 0; 
  $$ = nolink; 
}
;

%%

/**
 * Read rules from a stream (yyin) and build the corresponding compiler.
 * 
 * yyin defaults to stdin.
 */
CGCompiler * buildCompiler(istream* yyin = 0, Generator * g = 0)
{
  generator = g;
  delete lexer;
  delimiterlist = new taglist<int>();
  preferredlist = new taglist<int>();
  sets = new map<string,Set*>();
  corrections = new list<Correction *>();
  constraintsections = new list<list<Constraint*>*>();
  mappings = new list<Mapping*>();
  
  tag_register->add(*startTag);
  tag_register->add(*endTag);
  
  lexer = new cgrulesFlexLexer(yyin);
  yyparse();
  return new CGCompiler(delimiterlist, preferredlist, sets, 
			corrections, constraintsections, mappings);
}
